----------------------------------------------------------------------------------
-- Gibbous' Minmax Meters : Part of the GibsUI Package
-- Version : 0.3
-- Author : Gibbous <Crimson Knights>
-- Contact : Rockman922@hotmail.com
-- Environment : RIFT patch 1.5

-- the purpose of this addon is to enable complex minmaxing by recording
-- combat performance and displaying it in an easy-to-read manner.

-- please do contact in the event of bugs, or feature requests, either at the
-- above email, or on RiftUI.com
----------------------------------------------------------------------------------
-- CURRENT FEATURE LIST
-- > tracks multiple targets' DPS, Damage Done, and Percent contribution
-- > listing unit damage by damage type (as well as all contribution info)
-- > listing unit damage by ability (as well as all contribution info)
-- > displays this information with a single bar for each category
-- > click on bar segments to get a detailed readout of the target's performance
-- > saves UI state and tracking state between sessions
-- > moveable frame (thanks to Marru and Gumaden for the basecode)
-- > saves UI Position across sessions by character
-- > visiblity toggling as a button, rather than just a slash command
----------------------------------------------------------------------------------
-- PLANNED FEATURE LIST
-- > support incidental pets like Empty the Crypts
-- > expand data listed to include damage mitigated, and missed attacks
-- > track relevant tanking statistics, including blocks/parries/dodges per minute
-- > track relevant healing statistics, including overhealing and mana efficiency
-- > easy toggle between UI display modes
-- > alternate compact UI mode (for when space is more important than information
-- > performance boosts
-- > saving multiple encounters (by name of first target and time)
-- > track buff and debuff uptime, as well as their contribution to the damage
-- > track buff and debuff uptime, as well as their contribution to the healing
-- > track buff and debuff uptime, as well as their contribution to the defense
----------------------------------------------------------------------------------
-- KNOWN BUGS	
-- > tooltips are sometimes hidden by adjacent bars
-- > bar positions are slightly off
----------------------------------------------------------------------------------

-- Globals Begin
GibsUI = {
	verbose = false,
	loggedInToon = nil,
	toonPet = nil
}
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- SAVEDVARIABLES INFORMATION ONLY
-- default values
GibsUI_MMMGUI_Open = false -- to store
GibsUI_MMMGUI_Enabled = true -- it tracks in the background, so this can toggle it on and off
GibsUI_MMMGUI_Position = {}
GibsUI_MMMGUI_ButtonPos = {}
function savedVariablesLoad()
	if not GibbsUI_MMMGUI_Position then
		GibsUI_damageTracker.GUI.x = GibsUI_MMMGUI_Position.x
		GibsUI_damageTracker.GUI.y = GibsUI_MMMGUI_Position.y
	end
	GibsUI_damageTracker.control:SetPoint("TOPLEFT", UIParent, "TOPLEFT", GibsUI_damageTracker.GUI.x, GibsUI_damageTracker.GUI.y)
	
	if not GibsUI_MMM_GUI_ButtonPos then
		GibsUI_damageTracker.GUI.buttonx = GibsUI_MMMGUI_ButtonPos.x
		GibsUI_damageTracker.GUI.buttony = GibsUI_MMMGUI_ButtonPos.y
	end
	GibsUI_damageTracker.button:SetPoint("TOPLEFT", UIParent, "TOPLEFT", GibsUI_damageTracker.GUI.buttonx, GibsUI_damageTracker.GUI.buttony)
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- other GibsMMM specific variables
GibsUI_damageTracker = {
	versionNumber = 0.3,
	startTime = nil,
	control = nil,
	button = nil,
	combatTime = 0,
	totalDamageByAll = 0,
	tracking = false,
	trackedNames = {},
	trackedAbilities = {},
	damageHash = {},
	damageTypes = { "fire", "water", "earth", "life", "death", "air", "physical" },
	context = nil,
	
	-- GUI specific
	GUI = {
		mode = 0, -- 0 = GUI DPS mode
		selected = 1,
		typeSelected = 1,
		abilitySelected = 1,
		hover = false,
		typeHover = false,
		abilityHover = false,
		align = "TOPLEFT",
		x = 170,
		y = 170,
		buttonx = 170,
		buttony = 170 + 278,
		width = 500,
		height = 278,
		alpha = 0.5,
		meters = {}
	}
}
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- create the top level context for all GUI elements
GibsUI_damageTracker.context = UI.CreateContext("GibsMMM")
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- create the base of the damage meter
local BaseFrame = UI.CreateFrame("Frame", "Main", GibsUI_damageTracker.context)
BaseFrame:SetWidth(GibsUI_damageTracker.GUI.width)
BaseFrame:SetHeight(GibsUI_damageTracker.GUI.height)
BaseFrame:SetPoint("TOPLEFT", UIParent, "TOPLEFT")
BaseFrame:SetBackgroundColor(0,0,0,GibsUI_damageTracker.GUI.alpha)

if GibsUI_MMMGUI_Open == false then
	BaseFrame:SetVisible(false)
end

local ToggleButton = UI.CreateFrame("Texture", "TextureButton", GibsUI_damageTracker.context)
ToggleButton:SetTexture("GibsMMM","m3button.png")
ToggleButton:ResizeToTexture()
ToggleButton:SetPoint("TOPLEFT", UIParent, "TOPLEFT")
BaseFrame:SetBackgroundColor(0,0,0,GibsUI_damageTracker.GUI.alpha)

----------------------------------------------------------------------------------
-- mousemove on base bar (thanks to Marru and Gumaden for the basecode)
function BaseFrame.Event:RightDown()
        self.MouseDown = true
        mouseData = Inspect.Mouse()
        self.MyStartX = BaseFrame:GetLeft()
        self.MyStartY = BaseFrame:GetTop()
        self.StartX = mouseData.x - self.MyStartX
        self.StartY = mouseData.y - self.MyStartY
        tempX = BaseFrame:GetLeft()
        tempY = BaseFrame:GetTop()
        tempW = BaseFrame:GetWidth()
        tempH = BaseFrame:GetHeight()
        BaseFrame:ClearAll()
        BaseFrame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tempX, tempY)
        BaseFrame:SetWidth(tempW)
        BaseFrame:SetHeight(tempH)
        self:SetBackgroundColor(0.2,0.2,0.6,1)
end
function BaseFrame.Event:MouseMove()
        if self.MouseDown then
                local newX, newY
                mouseData = Inspect.Mouse()
                newX = mouseData.x - self.StartX
                newY = mouseData.y - self.StartY
                BaseFrame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", newX, newY)
        end
end
function BaseFrame.Event:RightUp()
        if self.MouseDown then
            GibsUI_MMMGUI_Position.x = math.floor(BaseFrame:GetLeft())
            GibsUI_MMMGUI_Position.y = math.floor(BaseFrame:GetTop())
            self.MouseDown = false
            self:SetBackgroundColor(0, 0, 0,GibsUI_damageTracker.GUI.alpha)
        end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- mousemove on button (thanks to Marru and Gumaden for the basecode)
function ToggleButton.Event:RightDown()
        self.MouseDown = true
        mouseData = Inspect.Mouse()
        self.MyStartX = ToggleButton:GetLeft()
        self.MyStartY = ToggleButton:GetTop()
        self.StartX = mouseData.x - self.MyStartX
        self.StartY = mouseData.y - self.MyStartY
        tempX = ToggleButton:GetLeft()
        tempY = ToggleButton:GetTop()
        ToggleButton:ClearAll()
        ToggleButton:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tempX, tempY)
        self:SetBackgroundColor(0.2,0.2,0.6,1)
end
function ToggleButton.Event:MouseMove()
        if self.MouseDown then
                local newX, newY
                mouseData = Inspect.Mouse()
                newX = mouseData.x - self.StartX
                newY = mouseData.y - self.StartY
                ToggleButton:SetPoint("TOPLEFT", UIParent, "TOPLEFT", newX, newY)
        end
end
function ToggleButton.Event:RightUp()
    if self.MouseDown then
        GibsUI_MMMGUI_ButtonPos.x = math.floor(ToggleButton:GetLeft())
        GibsUI_MMMGUI_ButtonPos.y = math.floor(ToggleButton:GetTop())
        self.MouseDown = false
        self:SetBackgroundColor(0, 0, 0,GibsUI_damageTracker.GUI.alpha)
    end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
GibsUI_damageTracker.control = BaseFrame -- STORED FOR FUTURE USE
GibsUI_damageTracker.button = ToggleButton
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
math.randomseed(Inspect.Time.Frame())
----------------------------------------------------------------------------------
-- clearHash : void
-- returns : void
-- zeros out the damage hash to prepare for fresh input
local function clearHash()
	for _, n in pairs (GibsUI_damageTracker.trackedNames) do
		local total = 0
		for _, d in pairs(GibsUI_damageTracker.damageTypes) do
			GibsUI_damageTracker.damageHash[n..d] = 0
			GibsUI_damageTracker.damageHash[n..d.."DPS"] = 0
		end
		GibsUI_damageTracker.damageHash[n.."total"] = 0
		GibsUI_damageTracker.damageHash[n.."DPS"] = 0
	end
	
	for _, a in pairs (GibsUI_damageTracker.trackedAbilities) do
		GibsUI_damageTracker.damageHash[a] = 0
		GibsUI_damageTracker.damageHash[a.."DPS"] = 0
	end
	
	if GibsUI.verbose then
		print ("Damage hash emptied")
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- isUnitTracked : string
-- returns : boolean
-- returns whether or not the unitName is found in the tracking table
local function isUnitTracked(unitName)
	for _, t in pairs(GibsUI_damageTracker.trackedNames) do
		if t == unitName then
			return true
		end
	end
	return false
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- isAbilityTracked : string
-- returns : boolean
-- returns whether or not the abilityName is found in the tracking table
local function isAbilityTracked(abilityName)
	for _, t in pairs(GibsUI_damageTracker.trackedAbilities) do
		if t == abilityName then
			return true
		end
	end
	return false
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- untrack : string
-- return : void
-- find and remove the unit from the tracking system
local function untrack(unit)
	local trackedSize = # GibsUI_damageTracker.trackedNames -- number of tracked units
	local removeIndex = -1 -- default
	for i = 0, trackedSize, 1 do -- loop
		if GibsUI_damageTracker.trackedNames[i] == unit then -- if it's a match
			removeIndex = i -- store it
			break -- we're done here
		end
	end
	if removeIndex > -1 then -- if we found it
		table.remove(GibsUI_damageTracker.trackedNames, removeIndex) -- remove it
		
		if GibsUI.verbose then 	-- verbosity check
			print (unit.." is no longer tracked by GibsMMM") -- debug
		end
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- track : string
-- returns : void
-- stores the name of the unit so that the system can track it
local function track(unit)
	table.insert(GibsUI_damageTracker.trackedNames, unit) -- insert the name into the table
	for _, e in pairs(GibsUI_damageTracker.damageTypes) do -- initialize damage hash for the target
		GibsUI_damageTracker.damageHash[unit..e] = 0
		GibsUI_damageTracker.damageHash[unit..e.."DPS"] = 0
	end
	GibsUI_damageTracker.damageHash[unit.."total"] = 0 -- finish
	GibsUI_damageTracker.damageHash[unit.."DPS"] = 0 -- done initializing
	
	if GibsUI.verbose then -- verbosity check
		print ("Damage hash for "..unit.." initialized") -- debug
		print (unit.." is now being tracked by GibsMMM") -- debug
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- startTracker : void
-- returns : void
-- enables the tracker when combat starts
-- ! currently works on Event.System.Secure.Enter, if Event.Combat.Enter or similar is created
-- switch to that for readability !
-- only works if enabled
local function startTracker()
	if not GibsUI_MMMGUI_Enabled then -- if it's not enabled
		GibsUI_damageTracker.tracking = false -- ensure it is disabled
		return -- get out of here
	end

	clearHash() -- clear the hash for a fresh run
	GibsUI_damageTracker.totalDamageByAll = 0 -- no damage done yet
	GibsUI_damageTracker.startTime = Inspect.Time.Frame() -- get the start time
	GibsUI_damageTracker.tracking = true -- start tracking
	
	if GibsUI.verbose then -- verbosity check
		print ("Entering Combat") -- debug
		print(GibsUI_damageTracker.startTime) -- debug
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- endTracker : void
-- returns : void
-- ends the tracker when out of combat
-- ! currently works on Event.System.Secure.Leave, if Event.Combat.Leave or similar is created
-- switch to that for readability !
local function endTracker()
	if GibsUI.verbose then -- verbosity check
		print ("Leaving Combat") -- debug
	end
	
	GibsUI_damageTracker.tracking = false -- disable
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- damageAccrue : Event.Combat.Damage params
-- return : void
-- processes damage events, adding them to the damage hash
-- only called when the tracker is active
local function damageAccrue(params)
	if not GibsUI_MMMGUI_Enabled then -- if it's not enabled
		GibsUI_damageTracker.tracking = false -- ensure it is disabled
		return -- get out of here
	end

	if not GibsUI_damageTracker.tracking then -- active check
		return
	end

	local caster = Inspect.Unit.Detail(params.caster) -- grab the caster
	local target = Inspect.Unit.Detail(params.target) -- grab the target
	
	if target == caster then -- don't count self-damage as DPS
		return
	end
	
	local damageType = "physical" -- default to physical damage type
	local damage = 0 -- default damage to zero
	
	if params.type then damageType = params.type end -- if the type exists (it's elemental) set it
	if params.damage then damage = params.damage end -- if damage was done, take it
	
	if caster and target then -- if both caster and target exist (bug prevention)
		if isUnitTracked(caster.name) then -- if we actually care about this caster			
			if not GibsUI_damageTracker.damageHash[caster.name..damageType] then -- make sure the hash exists
				GibsUI_damageTracker.damageHash[caster.name..damageType] = 0 -- if it doesn't, create it
			end
			if not GibsUI_damageTracker.damageHash[caster.name.."total"] then -- ditto here
				GibsUI_damageTracker.damageHash[caster.name.."total"] = 0 -- ditto here
			end
			GibsUI_damageTracker.damageHash[caster.name..damageType] = GibsUI_damageTracker.damageHash[caster.name..damageType] + damage -- add the damage to the proper target and damage type key
			GibsUI_damageTracker.damageHash[caster.name.."total"] = GibsUI_damageTracker.damageHash[caster.name.."total"] + damage -- add the damage to the proper target total key
			
			if not isAbilityTracked(caster.name..params.abilityName) then
				GibsUI_damageTracker.damageHash[caster.name..params.abilityName] = 0 -- initialize the damage hash for the ability
				GibsUI_damageTracker.damageHash[caster.name..params.abilityName.."DPS"] = 0
				table.insert(GibsUI_damageTracker.trackedAbilities, caster.name..params.abilityName) -- track the ability by caster name
			end
			GibsUI_damageTracker.damageHash[caster.name..params.abilityName] = GibsUI_damageTracker.damageHash[caster.name..params.abilityName] + damage -- add the damage up
		
			GibsUI_damageTracker.totalDamageByAll = GibsUI_damageTracker.totalDamageByAll + damage -- total tracked damage
			
			if GibsUI.verbose then -- verbosity check
				print(caster.name.." dealt "..damage.." "..damageType.." damage with "..params.abilityName.." to "..target.name) -- combat log style printout
				print("Total damage :"..GibsUI_damageTracker.totalDamageByAll) -- ditto
				print(caster.name..params.abilityName.." "..GibsUI_damageTracker.damageHash[caster.name..params.abilityName]) -- ditto
			end
		end
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- damageByUnit : void
-- returns : void
-- handles the processing of the DPS mode GUI for unit bars
local function damageByUnit()
	-- first, the total damage done and DPS portion
	local numTracked = 0 -- default
	GibsUI_damageTracker.GUI.hover = false -- preemptively dismiss tooltip
	
	if GibsUI_damageTracker.trackedNames then -- if there's tracked names
		numTracked = # GibsUI_damageTracker.trackedNames -- get them
	else -- if there's none, there's nothing we can do, so return
		return -- preemptive bug fix
	end
	
	if not GibsUI_damageTracker.GUI.meters["dpsBG"] then -- if the meter BG is not initialized
		local BGFrame = GibsUI_damageTracker.GUI.meters["bg"] -- create it
		local dpsBG = UI.CreateFrame("Frame", "dpsBG", BGFrame)
		dpsBG:SetPoint("TOPLEFT", BGFrame, "TOPLEFT", 15, 15)
		dpsBG:SetWidth(GibsUI_damageTracker.GUI.width - 30)
		dpsBG:SetHeight(72)
		dpsBG:SetBackgroundColor(0,0,0,0.5)
		dpsBG:SetLayer(1)
		GibsUI_damageTracker.GUI.meters["dpsBG"] = dpsBG -- store it
	end
	
	-- BEGIN TOTAL DAMAGE METER CODE
	local fullWidth = GibsUI_damageTracker.GUI.width - 36 -- full active meter width
	local pWidth = 0 -- default width, will change
	local nextMeterAt = 2 -- default point where the meter is going to appear (used to place side by side properly)
	local lr, lg, lb, la = 1, 1, 1, 1-- last colors (for bar random purposes) default white
	
	for _, n in pairs(GibsUI_damageTracker.trackedNames) do -- for every tracked name (need one bar for each of course)
		if not GibsUI_damageTracker.GUI.meters[n.."DPS"] then -- if that name doesn't have a bar for it yet
			local tempDPS = UI.CreateFrame("Frame", n.."TotalDPS", GibsUI_damageTracker.GUI.meters["dpsBG"]) -- create one!
			tempDPS:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["dpsBG"], "TOPLEFT", nextMeterAt, 2 ) -- parent to the background
			tempDPS:SetHeight(34) -- blah
			tempDPS:SetLayer(2) -- etc...
			-- class specific colors for your character only
			if n == GibsUI.loggedInToon.name then
				if GibsUI.loggedInToon.calling == "mage" then
					tempDPS:SetBackgroundColor(0.5, 0, 0.5,1) -- purple
				elseif GibsUI.loggedInToon.calling == "warrior" then
					tempDPS:SetBackgroundColor(0.7, 0, 0,1) -- red
				elseif GibsUI.loggedInToon.calling == "rogue" then
					tempDPS:SetBackgroundColor(0.7, 0.7, 0,1) -- yellow
				elseif GibsUI.loggedInToon.calling == "cleric" then
					tempDPS:SetBackgroundColor(0, 0.7, 0,1) -- green
				end
			else
				tempDPS:SetBackgroundColor(lr + math.random(-0.2, 0.2), lg + math.random(-0.2, 0.2), lb + math.random(-0.2, 0.2), la) -- random for everything else
			end
			lr, lg, lb, la = tempDPS:GetBackgroundColor()
			
			-- give the bars a left click function to select whose dps to watch
			tempDPS.Event.LeftClick = function() 
				local size = # GibsUI_damageTracker.trackedNames -- get the number of names
				for check = 0, size, 1 do -- loop
					if n == GibsUI_damageTracker.trackedNames[check] then -- if it matches
						GibsUI_damageTracker.GUI.selected = check -- that's the one we want to display the info for
						break -- we can leave now
					end
				end
			end
			
			-- give the bars a mouse over function to peek at who the bar represents
			tempDPS.Event.MouseIn = function() 
				GibsUI_damageTracker.GUI.hover = true -- you're hovering, make it so
				
				if not GibsUI_damageTracker.GUI.meters["hovertext"] then -- if the hovertext is nil (probably haven't hovered yet)
					local tempText = UI.CreateFrame("Text", "hoverText", GibsUI_damageTracker.GUI.meters[n.."DPS"]) -- create it
					tempText:SetPoint("CENTER", GibsUI_damageTracker.GUI.meters[n.."DPS"], "CENTER", 0, 0) -- defaults
					tempText:SetFontSize(15) -- won't change
					tempText:SetLayer(1000)
					tempText:SetBackgroundColor(0,0,0,0.5) -- won't change
					GibsUI_damageTracker.GUI.meters["hovertext"] = tempText
				end
				-- set the dynamic settings
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetParent(GibsUI_damageTracker.GUI.meters[n.."DPS"]) -- parent to the hovered bar
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetText(n) -- text is the name
				GibsUI_damageTracker.GUI.meters["hovertext"]:ResizeToText() -- resize as needed
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetPoint("CENTER", GibsUI_damageTracker.GUI.meters["hovertext"]:GetParent(), "CENTER", 0, 0) -- reposition
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetVisible(true) -- show it
			end
			
			-- give the bars a mouse out function to hide the tooltip when it's not needed
			tempDPS.Event.MouseOut = function()
				if not GibsUI_damageTracker.GUI.hover then -- if we hovered out, but another bar isn't being hovered over
					GibsUI_damageTracker.GUI.meters["hovertext"]:SetVisible(false) -- no more tooltip
				end
			end
			
			GibsUI_damageTracker.GUI.meters[n.."DPS"] = tempDPS	-- store the bar
		end
		
		-- set the meter's position
		-- next meter at determines the left (to ensure they stack up right)
		GibsUI_damageTracker.GUI.meters[n.."DPS"]:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["dpsBG"], "TOPLEFT", nextMeterAt, 2)
		local fullWidth = GibsUI_damageTracker.GUI.width - 36 -- locally store the full width the active bars can take
		local pWidth -- temp variable

		if GibsUI_damageTracker.totalDamageByAll == 0 then -- if there's no damage done yet
			pWidth = fullWidth/numTracked -- just give every bar an equal size
		else -- if we're in combat, or there was combat
			pWidth = (fullWidth * GibsUI_damageTracker.damageHash[n.."total"])/GibsUI_damageTracker.totalDamageByAll -- set each bar to the percent of damage done
		end
		GibsUI_damageTracker.GUI.meters[n.."DPS"]:SetWidth(pWidth) -- set
		
		if GibsUI.verbose then -- verbosity check
			print(n.." width : "..pWidth) -- debug the width
		end
		
		nextMeterAt = nextMeterAt + pWidth -- advance the next position to the end of this one
	end
	-- END TOTAL DAMAGE METER CODE
	
	-- BEGIN TOTAL DAMAGE TEXT READOUT
	if not GibsUI_damageTracker.GUI.meters["selectedColor"] then -- if the little color indicator is not created
		local tempC = UI.CreateFrame("Frame", "selColor", GibsUI_damageTracker.GUI.meters["dpsBG"]) -- create it
		tempC:SetHeight(30)
		tempC:SetWidth(30)
		tempC:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["dpsBG"], "TOPLEFT", 2, 40)
		GibsUI_damageTracker.GUI.meters["selectedColor"] = tempC -- store it
	end
	
	local selectedUnit = GibsUI_damageTracker.trackedNames[GibsUI_damageTracker.GUI.selected] -- store the selected unit
	if selectedUnit then
		local r, g, b, a = GibsUI_damageTracker.GUI.meters[selectedUnit.."DPS"]:GetBackgroundColor() -- store the associated bar color
		GibsUI_damageTracker.GUI.meters["selectedColor"]:SetBackgroundColor(r, g, b, a) -- set the cube to match the bar color'
		
		if not GibsUI_damageTracker.GUI.meters["infoReadout"] then -- if the info text is not created
			local tempR = UI.CreateFrame("Text", "info", GibsUI_damageTracker.GUI.meters["dpsBG"]) -- create it
			tempR:SetHeight(30)
			tempR:SetWidth(400)
			tempR:SetFontSize(15)
			tempR:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["dpsBG"], "TOPLEFT", 44, 42)
			GibsUI_damageTracker.GUI.meters["infoReadout"] = tempR -- store it
		end
		
		local percentDamage = 0 -- default local
		
		if GibsUI_damageTracker.totalDamageByAll > 0 then -- if damage has been done, we'll track stats
			percentDamage = ((100 * GibsUI_damageTracker.damageHash[selectedUnit.."total"]) / GibsUI_damageTracker.totalDamageByAll) -- calculate percent damage of the selected unit
		end
		
		-- display the damage readout of the selected unit
		GibsUI_damageTracker.GUI.meters["infoReadout"]:SetText(selectedUnit..": "
			..GibsUI_damageTracker.damageHash[selectedUnit.."total"].." Damage Done | " -- overall
			..math.floor(GibsUI_damageTracker.damageHash[selectedUnit.."DPS"]).." DPS | " -- damage per second
			..math.floor(percentDamage).."%") -- percent contribution
		-- END TOTAL DAMAGE TEXT READOUT
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- damageByType : void
-- returns : void
-- handles the processing of the DPS mode GUI for damage type bars
local function damageByType()
	local selectedUnit -- the name of the unit whose damage types we're tracking
	if GibsUI_damageTracker.trackedNames[GibsUI_damageTracker.GUI.selected] then -- if it exists
		selectedUnit = GibsUI_damageTracker.trackedNames[GibsUI_damageTracker.GUI.selected] -- set it
	else -- if there's no selected unit
		return -- move along, there's nothing we can do here
	end
	
	-- DAMAGE BY ELEMENT TYPE BEGIN
	if not GibsUI_damageTracker.GUI.meters["typeBG"] then -- if the meter's BG is not initialized
		local BGFrame = GibsUI_damageTracker.GUI.meters["bg"] -- create it
		local typeBG = UI.CreateFrame("Frame", "typeBG", BGFrame)
		typeBG:SetPoint("TOPLEFT", BGFrame, "TOPLEFT", 15, 102)
		typeBG:SetWidth(GibsUI_damageTracker.GUI.width - 30)
		typeBG:SetHeight(72)
		typeBG:SetBackgroundColor(0,0,0,0.5)
		typeBG:SetLayer(1)
		GibsUI_damageTracker.GUI.meters["typeBG"] = typeBG -- store it
	end
	
	local pWidth = 0 -- reset this variable for reuse
	local nextMeterAt = 2 -- this one as well
	local numTracked = # GibsUI_damageTracker.damageTypes -- get the number of possible bars
	local fullWidth = GibsUI_damageTracker.GUI.width - 36 -- full active meter width
	
	for _, n in pairs(GibsUI_damageTracker.damageTypes) do -- for every tracked damage type (need one bar for each of course)
		if not GibsUI_damageTracker.GUI.meters[n] then -- if that type doesn't have a bar for it yet
			local tempDPS = UI.CreateFrame("Frame", n, GibsUI_damageTracker.GUI.meters["typeBG"]) -- create one!
			tempDPS:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["typeBG"], "TOPLEFT", nextMeterAt, 2) -- parent to the background
			tempDPS:SetHeight(34) -- blah
			tempDPS:SetLayer(2) -- etc...
			-- specific colors by element for the bars
			if n == "fire" then
				tempDPS:SetBackgroundColor(0.7, 0, 0,1)
			elseif n == "water"  then
				tempDPS:SetBackgroundColor(0, 0, 0.7,1)
			elseif n == "earth"  then
				tempDPS:SetBackgroundColor(0.7, 0.7, 0,1) 
			elseif n == "air"  then
				tempDPS:SetBackgroundColor(0.7, 0.7, 0.7, 1)
			elseif n == "life" then
				tempDPS:SetBackgroundColor(0, 0.7, 0,1)
			elseif n == "death"  then
				tempDPS:SetBackgroundColor(0.5, 0, 0.5,1)
			else
				tempDPS:SetBackgroundColor(0.2, 0.2, 0.2,1)
			end
			
			-- give the bars a left click function to select what element to watch
			tempDPS.Event.LeftClick = function() 
				local size = # GibsUI_damageTracker.damageTypes -- get the number of elements
				for check = 0, size, 1 do -- loop
					if n == GibsUI_damageTracker.damageTypes[check] then -- if it matches
						GibsUI_damageTracker.GUI.typeSelected = check -- that's the one we want to display the info for
						break -- we can leave now
					end
				end
			end
			
			-- give the bars a mouse over function to peek at who the bar represents
			tempDPS.Event.MouseIn = function() 
				GibsUI_damageTracker.GUI.hover = true -- you're hovering, make it so
				
				if not GibsUI_damageTracker.GUI.meters["hovertext"] then -- if the hovertext is nil (probably haven't hovered yet)
					local tempText = UI.CreateFrame("Text", "hoverText", GibsUI_damageTracker.GUI.meters[n]) -- create it
					tempText:SetPoint("CENTER", GibsUI_damageTracker.GUI.meters[n], "CENTER", 0, 0) -- defaults
					tempText:SetFontSize(15) -- won't change
					tempText:SetLayer(1000)
					tempText:SetBackgroundColor(0,0,0,0.5) -- won't change
					GibsUI_damageTracker.GUI.meters["hovertext"] = tempText
				end
				-- set the dynamic settings
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetParent(GibsUI_damageTracker.GUI.meters[n]) -- parent to the hovered bar
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetText(n) -- text is the name
				GibsUI_damageTracker.GUI.meters["hovertext"]:ResizeToText() -- resize as needed
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetPoint("CENTER", GibsUI_damageTracker.GUI.meters["hovertext"]:GetParent(), "CENTER", 0, 0) -- reposition
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetVisible(true) -- show it
			end
			
			-- give the bars a mouse out function to hide the tooltip when it's not needed
			tempDPS.Event.MouseOut = function()
				if not GibsUI_damageTracker.GUI.hover then -- if we hovered out, but another bar isn't being hovered over
					GibsUI_damageTracker.GUI.meters["hovertext"]:SetVisible(false) -- no more tooltip
				end
			end
			
			GibsUI_damageTracker.GUI.meters[n] = tempDPS	-- store the bar
		end
		
		-- set the meter's position
		-- next meter at determines the left (to ensure they stack up right)
		GibsUI_damageTracker.GUI.meters[n]:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["typeBG"], "TOPLEFT", nextMeterAt, 2)

		if GibsUI_damageTracker.damageHash[selectedUnit.."total"] == 0 then -- if there's no damage done yet
			pWidth = fullWidth/numTracked -- just give every bar an equal size
		else -- if we're in combat, or there was combat
			pWidth = (fullWidth * GibsUI_damageTracker.damageHash[selectedUnit..n])/GibsUI_damageTracker.damageHash[selectedUnit.."total"] -- set each bar to the percent of damage done
		end
		GibsUI_damageTracker.GUI.meters[n]:SetWidth(pWidth) -- set
		
		if GibsUI.verbose then -- verbosity check
			print(selectedUnit..".."..n.." width : "..pWidth) -- debug the width
		end
		
		nextMeterAt = nextMeterAt + pWidth -- advance the next position to the end of this one
	end
	-- END TOTAL DAMAGE METER CODE
	
	if not GibsUI_damageTracker.GUI.meters["typeSelectedColor"] then -- if the little color indicator is not created
		local tempC = UI.CreateFrame("Frame", "typeSelColor", GibsUI_damageTracker.GUI.meters["typeBG"]) -- create it
		tempC:SetHeight(30)
		tempC:SetWidth(30)
		tempC:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["typeBG"], "TOPLEFT", 2, 40)
		GibsUI_damageTracker.GUI.meters["typeSelectedColor"] = tempC -- store it
	end
	
	local selectedType = GibsUI_damageTracker.damageTypes[GibsUI_damageTracker.GUI.typeSelected] -- store the selected type
	
	if selectedType then
		
		if GibsUI_damageTracker.damageHash[selectedUnit..selectedType] == 0 and GibsUI_damageTracker.damageHash[selectedUnit.."total"] > 0 then
			GibsUI_damageTracker.GUI.typeSelected = 1
			selectedType = GibsUI_damageTracker.damageTypes[GibsUI_damageTracker.GUI.typeSelected]
			while GibsUI_damageTracker.damageHash[selectedUnit..selectedType] == 0 do
				GibsUI_damageTracker.GUI.typeSelected = GibsUI_damageTracker.GUI.typeSelected + 1 -- this ensures that while you're in combat, it defaults to a damage type you're actually dealing
				selectedType = GibsUI_damageTracker.damageTypes[GibsUI_damageTracker.GUI.typeSelected]
			end
		end
		
		local r, g, b, a = GibsUI_damageTracker.GUI.meters[selectedType]:GetBackgroundColor() -- store the associated bar color
		GibsUI_damageTracker.GUI.meters["typeSelectedColor"]:SetBackgroundColor(r, g, b, a) -- set the cube to match the bar color'
		
		if not GibsUI_damageTracker.GUI.meters["typeInfoReadout"] then -- if the info text is not created
			local tempR = UI.CreateFrame("Text", "typeinfo", GibsUI_damageTracker.GUI.meters["typeBG"]) -- create it
			tempR:SetHeight(30)
			tempR:SetWidth(400)
			tempR:SetFontSize(15)
			tempR:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["typeBG"], "TOPLEFT", 44, 42)
			GibsUI_damageTracker.GUI.meters["typeInfoReadout"] = tempR -- store it
		end
		
		local percentDamage = 0 -- default local
		
		if GibsUI_damageTracker.damageHash[selectedUnit.."total"] > 0 then -- if damage has been done, we'll track stats
			percentDamage = ((100 * GibsUI_damageTracker.damageHash[selectedUnit..selectedType]) / GibsUI_damageTracker.damageHash[selectedUnit.."total"]) -- calculate percent damage of the selected unit
		end	
		
		-- display the damage readout of the selected unit
		GibsUI_damageTracker.GUI.meters["typeInfoReadout"]:SetText(selectedType..": "
			..GibsUI_damageTracker.damageHash[selectedUnit..selectedType].." Damage Done | " -- overall
			..math.floor(GibsUI_damageTracker.damageHash[selectedUnit..selectedType.."DPS"]).." DPS | " -- damage per second
			..math.floor(percentDamage).."%") -- percent contribution
		-- END TOTAL DAMAGE TEXT READOUT
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- damageByAbility : void
-- returns : void
-- handles the processing of the DPS mode GUI for ability damage bars
local function damageByAbility()
	local selectedUnit -- the name of the unit whose abilities we're tracking
	if GibsUI_damageTracker.trackedNames[GibsUI_damageTracker.GUI.selected] then -- if it exists
		selectedUnit = GibsUI_damageTracker.trackedNames[GibsUI_damageTracker.GUI.selected] -- set it
	else -- if there's no selected unit
		return -- move along, there's nothing we can do here
	end
		
	-- DAMAGE BY ABILITY BEGIN
	if not GibsUI_damageTracker.GUI.meters["abilityBG"] then -- if the meter's BG is not initialized
		local BGFrame = GibsUI_damageTracker.GUI.meters["bg"] 
		local abilityBG = UI.CreateFrame("Frame", "abilityBG", BGFrame) -- create it
		abilityBG:SetPoint("TOPLEFT", BGFrame, "TOPLEFT", 15, 189)
		abilityBG:SetWidth(GibsUI_damageTracker.GUI.width - 30)
		abilityBG:SetHeight(72)
		abilityBG:SetBackgroundColor(0,0,0,0.5)
		abilityBG:SetLayer(1)
		GibsUI_damageTracker.GUI.meters["abilityBG"] = abilityBG -- store it
	end
	
	if not GibsUI_damageTracker.GUI.meters["abilityBG2"] then -- if the meter's BG is not initialized
		local BGFrame = GibsUI_damageTracker.GUI.meters["abilityBG"] 
		abilityBG = UI.CreateFrame("Frame", "abilityBG2", BGFrame) -- create it
		abilityBG:SetPoint("TOPLEFT", BGFrame, "TOPLEFT", 2, 2)
		abilityBG:SetWidth(GibsUI_damageTracker.GUI.width - 36)
		abilityBG:SetHeight(34)
		abilityBG:SetBackgroundColor(1,1,1,0.1)
		abilityBG:SetLayer(1)
		GibsUI_damageTracker.GUI.meters["abilityBG2"] = abilityBG -- store it
	end
	
	local pWidth = 0 -- reset this variable for reuse
	local nextMeterAt = 2 -- this one as well
	local numTracked = # GibsUI_damageTracker.trackedAbilities -- get the number of possible bars
	local fullWidth = GibsUI_damageTracker.GUI.width - 36 -- full active meter width
	local lr, lg, lb, la = 0.5, 0.5,0.5, 1-- last colors (for bar random purposes) default white
	
	for _, n in pairs(GibsUI_damageTracker.trackedAbilities) do -- for every tracked ability (need one bar for each of course)
		if not GibsUI_damageTracker.GUI.meters[n] then -- if that ability doesn't have a bar for it yet
			local tempDPS = UI.CreateFrame("Frame", n, GibsUI_damageTracker.GUI.meters["abilityBG"]) -- create one!
			tempDPS:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["abilityBG"], "TOPLEFT", nextMeterAt, 2) -- parent to the background
			tempDPS:SetHeight(34) -- blah
			tempDPS:SetLayer(2) -- etc...
			tempDPS:SetBackgroundColor(lr + math.random(-0.2, 0.2), lg + math.random(-0.2, 0.2), lb + math.random(-0.2, 0.2), la)
			-- give the bars a left click function to select what element to watch
			tempDPS.Event.LeftClick = function() 
				local size = # GibsUI_damageTracker.trackedAbilities -- get the number of elements
				for check = 0, size, 1 do -- loop
					if n == GibsUI_damageTracker.trackedAbilities[check] then -- if it matches
						GibsUI_damageTracker.GUI.abilitySelected = check -- that's the one we want to display the info for
						break -- we can leave now
					end
				end
			end
			
			-- give the bars a mouse over function to peek at who the bar represents
			tempDPS.Event.MouseIn = function() 
				GibsUI_damageTracker.GUI.hover = true -- you're hovering, make it so
				
				local info = string.sub(n, string.len(GibsUI_damageTracker.trackedNames[GibsUI_damageTracker.GUI.selected])+1)
				if not GibsUI_damageTracker.GUI.meters["hovertext"] then -- if the hovertext is nil (probably haven't hovered yet)
					local tempText = UI.CreateFrame("Text", "hoverText", GibsUI_damageTracker.GUI.meters[n]) -- create it
					tempText:SetPoint("CENTER", GibsUI_damageTracker.GUI.meters[n], "CENTER", 0, 0) -- defaults
					tempText:SetFontSize(15) -- won't change
					tempText:SetBackgroundColor(0,0,0,0.5) -- won't change
					GibsUI_damageTracker.GUI.meters["hovertext"] = tempText
				end
				-- set the dynamic settings
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetParent(GibsUI_damageTracker.GUI.meters[n]) -- parent to the hovered bar
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetText(info) -- text is the name
				GibsUI_damageTracker.GUI.meters["hovertext"]:ResizeToText() -- resize as needed
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetPoint("CENTER", GibsUI_damageTracker.GUI.meters["hovertext"]:GetParent(), "CENTER", 0, 0) -- reposition
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetLayer(10)
				GibsUI_damageTracker.GUI.meters["hovertext"]:SetVisible(true) -- show it
			end
			
			-- give the bars a mouse out function to hide the tooltip when it's not needed
			tempDPS.Event.MouseOut = function()
				if not GibsUI_damageTracker.GUI.hover then -- if we hovered out, but another bar isn't being hovered over
					GibsUI_damageTracker.GUI.meters["hovertext"]:SetVisible(false) -- no more tooltip
				end
			end
			
			GibsUI_damageTracker.GUI.meters[n] = tempDPS	-- store the bar
		end
		lr, lg, lb, la = GibsUI_damageTracker.GUI.meters[n]:GetBackgroundColor()
		
		-- set the meter's position
		-- next meter at determines the left (to ensure they stack up right)
		GibsUI_damageTracker.GUI.meters[n]:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["abilityBG"], "TOPLEFT", nextMeterAt, 2)

		if GibsUI_damageTracker.damageHash[selectedUnit.."total"] == 0 then -- if there's no damage done yet
			pWidth = 0
		else
			if string.find(n, selectedUnit, 1) then -- if we're in combat, or there was combat and the ability hash exists
				pWidth = (fullWidth * GibsUI_damageTracker.damageHash[n])/GibsUI_damageTracker.damageHash[selectedUnit.."total"] -- set each bar to the percent of damage done
			else
				pWidth = 0
			end
		end
		GibsUI_damageTracker.GUI.meters[n]:SetWidth(pWidth) -- set
		
		if GibsUI.verbose then -- verbosity check
			print(n.." width : "..pWidth) -- debug the width
		end
		
		nextMeterAt = nextMeterAt + pWidth -- advance the next position to the end of this one
	end
	-- END TOTAL DAMAGE METER CODE
	
	if not GibsUI_damageTracker.GUI.meters["abilSelectedColor"] then -- if the little color indicator is not created
		local tempC = UI.CreateFrame("Frame", "abilSelColor", GibsUI_damageTracker.GUI.meters["abilityBG"]) -- create it
		tempC:SetHeight(30)
		tempC:SetWidth(30)
		tempC:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["abilityBG"], "TOPLEFT", 2, 40)
		tempC:SetBackgroundColor(1,1,1,0.1)
		GibsUI_damageTracker.GUI.meters["abilSelectedColor"] = tempC -- store it
	end
	
	if not GibsUI_damageTracker.GUI.meters["abilInfoReadout"] then -- if the info text is not created
		local tempR = UI.CreateFrame("Text", "abilinfo", GibsUI_damageTracker.GUI.meters["abilityBG"]) -- create it
		tempR:SetHeight(30)
		tempR:SetWidth(400)
		tempR:SetFontSize(15)
		tempR:SetText("No Ability: 0 Damage Done | 0 DPS | 0%")
		tempR:SetPoint("TOPLEFT", GibsUI_damageTracker.GUI.meters["abilityBG"], "TOPLEFT", 44, 42)
		GibsUI_damageTracker.GUI.meters["abilInfoReadout"] = tempR -- store it
	end
	
	local selectedAbility = GibsUI_damageTracker.trackedAbilities[GibsUI_damageTracker.GUI.abilitySelected] -- store the selected ability
		
	if selectedAbility then
		if not string.find(selectedAbility, selectedUnit, 1) or GibsUI_damageTracker.damageHash[selectedAbility] == 0 then -- if the ability isn't by this unit, or has 0 damage, look for another
			GibsUI_damageTracker.GUI.abilitySelected = 1 -- start at the top of the list
			local loops = 0
			local numTracked = # GibsUI_damageTracker.trackedAbilities -- get the nubmer we have stored
			
			while loops < numTracked do -- loop
				GibsUI_damageTracker.GUI.abilitySelected = GibsUI_damageTracker.GUI.abilitySelected + 1 -- set stuff
				selectedAbility = GibsUI_damageTracker.trackedAbilities[GibsUI_damageTracker.GUI.abilitySelected]
				if selectedAbility then -- if it's not nil
					if string.find(selectedAbility, selectedUnit, 1) then -- ifit's our ability
						if GibsUI_damageTracker.damageHash[selectedAbility] > 0 then -- if it's damaged
							break -- we're good, let's leave
						end
					end
				elseif loops == numTracked then -- else, if we're at the end of our rope
					selectedAbility = nil -- better make sure it's nil
				end
				loops = loops + 1 -- increment
			end
		end
		
		if selectedAbility then
			local r, g, b, a = GibsUI_damageTracker.GUI.meters[selectedAbility]:GetBackgroundColor() -- store the associated bar color
			GibsUI_damageTracker.GUI.meters["abilSelectedColor"]:SetBackgroundColor(r, g, b, a) -- set the cube to match the bar color'
		end
		
		local percentDamage = 0 -- default local
		local DPS = 0 -- default local
		local totalDone = 0 -- default local
		GibsUI_damageTracker.GUI.meters["abilInfoReadout"]:SetText("No Ability: 0 Damage Done | 0 DPS | 0%") -- default
		if GibsUI_damageTracker.damageHash[selectedUnit.."total"] > 0 and selectedAbility then
			if GibsUI_damageTracker.damageHash[selectedAbility] then -- if damage has been done, we'll track stats
				totalDone = GibsUI_damageTracker.damageHash[selectedAbility]
				percentDamage = ((100 * GibsUI_damageTracker.damageHash[selectedAbility]) / GibsUI_damageTracker.damageHash[selectedUnit.."total"]) -- calculate percent damage of the selected unit
			end	
					
			if GibsUI_damageTracker.damageHash[selectedAbility.."DPS"] then
				DPS = GibsUI_damageTracker.damageHash[selectedAbility.."DPS"]
			end
			local info = string.sub(selectedAbility, string.len(selectedUnit)+1)
			-- display the damage readout of the selected unit
			GibsUI_damageTracker.GUI.meters["abilInfoReadout"]:SetText(info..": "
				..totalDone.." Damage Done | " -- overall
				..math.floor(DPS).." DPS | " -- damage per second
				..math.floor(percentDamage).."%") -- percent contribution
			-- END TOTAL DAMAGE TEXT READOUT
		end				
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- tick : void
-- returns : void
-- powerhouse method, ensures that player and pet are tracked, and
-- processes time, as well as updating the UI when it is open
local function tick()
	local toon, pet
	if GibsUI.loggedInToon == nil then -- if the logged in toon is still blank
		GibsUI.loggedInToon = Inspect.Unit.Detail("player") -- get the logged in toon
		toon = GibsUI.loggedInToon
		
		if toon then
			track(toon.name) -- make sure the character is tracked
		end
	end
	
	if GibsUI.toonPet == nil then  -- if we don't have a pet
		GibsUI.toonPet = Inspect.Unit.Detail("player.pet") -- try to get the pet
		
		pet = GibsUI.toonPet
		if pet then -- if it exists, track it
			track(pet.name)
		end
	elseif Inspect.Unit.Detail("player.pet") == nil then -- if the pet is somehow gone now
		untrack(GibsUI.toonPet.name) -- remove it
		GibsUI.toonPet = nil -- nil it so we start searching for it again
	end
	
	if GibsUI.verbose then -- verbosity check
		toon = GibsUI.loggedInToon
		pet = GibsUI.toonPet
		print ("Initializing GibsMMM")
		name = toon.titlePrefix.." "..toon.name..toon.titleSuffix -- to look nice
		print ("logged in as "..name)  -- full name
		
		-- stats for whatever reason
		print ("HP: "..toon.health.."/"..toon.healthMax)  
		
		if toon.calling == "rogue" then
			print("E: "..toon.energy.."/100")
		elseif toon.calling == "warrior" then
			print("P: "..toon.power.."/100")
		elseif toon.calling == "cleric" then
			print("MP: "..toon.mana.."/"..toon.manaMax)
		elseif toon.calling == "mage" then
			print("MP: "..toon.mana.."/"..toon.manaMax)
			print("C: "..toon.charge.."/100")
		end
		
		if GibsUI.toonPet then -- print the pet too if there is one
			print ("Pet: "..pet.name)
		end
	end
		
	if GibsUI_damageTracker.tracking and GibsUI_MMMGUI_Enabled then -- if the tracker is active and enabled
		GibsUI_damageTracker.combatTime = Inspect.Time.Frame() - GibsUI_damageTracker.startTime -- update time in combat
		
		if GibsUI.verbose then -- verbosity check
			print("Time in Combat: "..GibsUI_damageTracker.combatTime) -- debug
		end
		
		for _, n in pairs(GibsUI_damageTracker.trackedNames) do -- for each tracked name
			GibsUI_damageTracker.damageHash[n.."DPS"] = (GibsUI_damageTracker.damageHash[n.."total"] / GibsUI_damageTracker.combatTime) -- update the DPS hash
			
			for _, t in pairs(GibsUI_damageTracker.damageTypes) do -- for each damage type
				GibsUI_damageTracker.damageHash[n..t.."DPS"] = (GibsUI_damageTracker.damageHash[n..t] / GibsUI_damageTracker.combatTime) -- update the DPS
				if GibsUI.verbose then -- verbosity check
					print(n.." current "..t.." DPS "..GibsUI_damageTracker.damageHash[n.."DPS"]) -- debug
				end
			end
			
			for _, a in pairs(GibsUI_damageTracker.trackedAbilities) do
				if GibsUI_damageTracker.damageHash[a.."DPS"] then
					GibsUI_damageTracker.damageHash[a.."DPS"] = (GibsUI_damageTracker.damageHash[a] / GibsUI_damageTracker.combatTime) -- update the DPS
					if GibsUI.verbose then -- verbosity check
						print(a.." DPS "..GibsUI_damageTracker.damageHash[a.."DPS"]) -- debug
					end			
				end
			end
			
			if GibsUI.verbose then -- verbosity check
				print(n.." current DPS "..GibsUI_damageTracker.damageHash[n.."DPS"]) -- debug
			end
		end	
	end
	
	if GibsUI_MMMGUI_Open then -- if the UI is open
		if not GibsUI_damageTracker.GUI.meters["bg"] then -- if the main frame is still not created
			GibsUI_damageTracker.GUI.meters["bg"] = GibsUI_damageTracker.control -- set so that it's not  multi-created (memory leak!)
		end
		GibsUI_damageTracker.GUI.meters["bg"]:SetVisible(true)
		if GibsUI_damageTracker.GUI.mode == 0 then -- mode 0 is DPS
			damageByUnit() -- unit tracking
			damageByType() -- type tracking
			damageByAbility() -- ability tracking
		end
	else -- if the UI isn't open
		if GibsUI_damageTracker.GUI.meters["bg"] then -- if we've allocated the GUI frame
 			GibsUI_damageTracker.GUI.meters["bg"]:SetVisible(false)
		end
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- printHash : void
-- returns : void
-- debug method, prints the entire damage hash
-- only available while GibsUI.verbose is true
local function printHash()	
	if not GibsUI.verbose then -- verbose check
		return
	end
	
	for _, n in pairs (GibsUI_damageTracker.trackedNames) do -- top loop cycles through tracked names
		local total = 0
		for _, d in pairs(GibsUI_damageTracker.damageTypes) do -- inner loop cycles through damage types
			local dam = 0
			if GibsUI_damageTracker.damageHash[n..d] then -- if it exists
				dam = GibsUI_damageTracker.damageHash[n..d] -- grab its value
			end
			print(n.." "..d.." "..dam) -- print, 0 default
		end
		
		for _, a in pairs(GibsUI_damageTracker.trackedAbilities) do
			local dam = 0
			if GibsUI_damageTracker.damageHash[a] and string.find(n, a) then
				dam = GibsUI_damageTracker.damageHash[a]
			end
			print (a.." "..dam)
		end
		
		print ("Total damage done by "..n..": "..GibsUI_damageTracker.damageHash[n.."total"]) -- print hash total 
		print ("Calculated DPS realtime for "..n..": "..GibsUI_damageTracker.damageHash[n.."DPS"]) -- print realtime DPS
	end
	print ("End of Hash") -- signify the end of printing
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- printTracked : void
-- returns : void
-- debug method that prints out the entire tracked table
-- only available while GibsUI.verbose is true
local function printTracked()
	if not GibsUI.verbose then
		return
	end
	
	for _, t in pairs(GibsUI_damageTracker.trackedNames) do
		print(t)
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- toggleUI : void
-- returns : void
-- toggles the GibsMMM UI to on and off (currently /toggleUI)
local function toggleUI()
	if GibsUI_MMMGUI_Open then -- if open, close
		GibsUI_MMMGUI_Open = false
	else -- if closed, open
		GibsUI_MMMGUI_Open = true
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- verboseToggle : void
-- returns : void
-- toggles verbosity on and off for debugging purposes
local function verboseToggle()
	if GibsUI.verbose then -- if it's on
		print("GibsMMM is no longer verbose") -- turn it off
		GibsUI.verbose = false
	else -- if it's off
		print("GibsMMM is now verbose") -- turn it on
		GibsUI.verbose = true
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- toggleEnabled : void
-- returns : void
-- toggles tracking on and off
local function toggleEnabled()
	if GibsUI_MMMGUI_Enabled then -- if it's on
		print("GibsMMM is now disabled") -- turn it off
		GibsUI_MMMGUI_Enabled = false
	else -- if it's off
		print("GibsMMM is now enabled") -- turn it on
		GibsUI_MMMGUI_Enabled = true
	end
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- showHelp : void
-- returns : void
-- prints the ui help
local function showHelp()
	print("Welcome to GibsMMM (M3) version "..GibsUI_damageTracker.versionNumber)
	print("/m3ToggleVerbose : toggles debug messages and functions")
	print("/m3PrintTracked : debug function - prints tracked units when verbose")
	print("/m3PrintHash : debug function - prints damage hash when verbose")
	print("/m3ToggleUI : toggle main UI visibility")
	print("/m3ToggleEnabled : toggle damage tracking")
	print("LEFT CLICK 	on the blue M3 button to toggle the UI")
	print("RIGHT CLICK and drag to move the UI frame or M3 Button")
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- init : void
-- returns : void
-- registers necessary events and slash commands
local function init() -- called at the bottom of the script
	table.insert(Event.System.Update.Begin, {tick, "GibsMMM", "refresh"})
	table.insert(Event.Combat.Damage, {damageAccrue, "GibsMMM", "combat_damage_info"})
	table.insert(Event.System.Secure.Enter, {startTracker, "GibsMMM", "starts the combat tracker"})
	table.insert(Event.System.Secure.Leave, {endTracker, "GibsMMM", "ends the combat tracker"})
	
	table.insert (Command.Slash.Register("m3ToggleVerbose"), {verboseToggle, "GibsMMM", "verboseToggle"})
	table.insert (Command.Slash.Register("m3PrintTracked"), {printTracked, "GibsMMM", "Print tracked units"})
	table.insert (Command.Slash.Register("m3PrintHash"), {printHash, "GibsMMM", "Print hashtable of damage"})
	table.insert (Command.Slash.Register("m3ToggleUI"), {toggleUI, "GibsMMM", "toggleUI"})
	table.insert (Command.Slash.Register("m3ToggleEnabled"), {toggleEnabled, "GibsMMM", "toggleEnable"})
	table.insert (Command.Slash.Register("gibsMMM"), {showHelp, "GibsMMM", "help"})
	
	table.insert(Event.Addon.SavedVariables.Load.End, {savedVariablesLoad, "GibsMMM", "save the saved stuff"})
	print("Type /gibsMMM for a list of commands and features")
end
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- button click toggle : void
-- returns void
-- handles the left-click UI toggle
function ToggleButton.Event:LeftClick()
	toggleUI()
end
----------------------------------------------------------------------------------

init()

